class BaseWindow(object):
    def __init__(self, size=(512, 512)):
        self.size = size
        self._behaviours = []

    def add_behaviours(self, behaviours):
        self._behaviours.extend(behaviours)
        return self

    def add_behaviour(self, behaviour):
        self._behaviours.append(behaviour)
        return self

    def show(self, title="Scene"):
        raise NotImplementedError()



from PIL import Image

from ..utils import save_frame
from . import Behaviour


class SaveFrames(Behaviour):
    """Save the rendered frames to a file.

    Arguments
    ---------
        path: Path to the file to save the frame. Python's format() will be
              used to add the index of the frame in the file name as follows
              '/path/to/frame_{}.png' -> /path/to/frame_0.png .
        every_n: int, Save every n frames instead of all frames.
    """
    def __init__(self, path, every_n=1):
        self._path = path
        self._every_n = every_n
        self._i = 0

    def behave(self, params):
        self._i += 1

        if (self._i % self._every_n) != 0:
            return

        frame = params.frame()
        path = self._path.format(self._i)
        save_frame(path, frame)


class SaveGif(Behaviour):
    """Save the rendered frames to a GIF animation.

    Arguments
    ---------
        path: Path to the file to save the gif animation
        every_n: int, save every n frames instead of all frames (default: 1)
        duration: int, the display duration in milliseconds per frame
                  (default: 16*every_n)
        loop: int, number of times the gif should loop (default: 0 which means
                   for ever)
        optimize: bool, whether to attempt to optimize the color pallete
                  (default: True)
    """
    def __init__(self, path, every_n=1, duration=None, loop=0, optimize=True):
        self._path = path
        self._every_n = every_n
        self._duration = duration or 16*every_n
        self._loop = loop

        self._i = 0
        self._images = []

    def behave(self, params):
        self._i += 1

        if not params.last_call and (self._i % self._every_n) != 0:
            return

        frame = params.frame()
        self._images.append(Image.fromarray(frame))

        if params.last_call:
            self._images[0].save(
                self._path,
                save_all=True,
                append_images=self._images[1:],
                duration=self._duration,
                loop=self._loop
            )
            self._i = 0
            self._images = []



from collections import defaultdict
import os

import numpy as np

from ..renderables.material import Material
from . import read_material_file
from .material import MaterialReader
from .utils import get_file, close_file


class MultiMaterialObjReader:
    def __init__(self, filepath, material_filepath=None, material_ext=None,
                 color=(0.5, 0.5, 0.5)):
        self.default_material = self.make_default_material(filepath, color)
        self.materials = (
            self.read_materials(filepath, material_filepath, material_ext)
        )
        self.objects = self.read_objects(filepath)

    def make_default_material(self, filepath, color):
        # extract the filename from a possible file-like python object
        filename = filepath.name if hasattr(filepath, "name") else filepath
        texture_path = os.path.join(os.path.dirname(filename), "texture.png")
        if os.path.exists(texture_path):
            return Material.with_texture_image(texture_path, diffuse=color)
        else:
            return Material(diffuse=color)

    def read_materials(self, filepath, material_filepath, material_ext):
        try:
            # Read the material file from the obj file
            f = get_file(filepath)
            try:
                material = [
                    l.split()[1:][0] for l in f if l.startswith("mtllib")
                ][0]
                filename = f.name if hasattr(f, "name") else filepath
                material = os.path.join(os.path.dirname(filename), material)

            # or use the provided material filepath
            except IndexError:
                material = material_filepath
        finally:
            close_file(f, filepath)

        if material is not None and os.path.exists(material):
            return read_material_file(material, ext=material_ext)
        else:
            return MaterialReader()

    def read_objects(self, filepath):
        try:
            f = get_file(filepath)
            lines = f.readlines()

            # Collect all the vertices, normals and uv coordinates
            vertices_all = np.array([
                list(map(float, l.strip().split()[1:4]))
                for l in lines if l.startswith("v ")
            ], dtype=np.float32)
            normals_all = np.array([
                list(map(float, l.strip().split()[1:]))
                for l in lines if l.startswith("vn ")
            ])
            uv_all = np.array([
                list(map(float, l.strip().split()[1:]))
                for l in lines if l.startswith("vt ")
            ])

            # Gather vertices, normals and uv for each face and group them by
            # material
            material = None
            vertices = defaultdict(list)
            normals = defaultdict(list)
            uv = defaultdict(list)
            for line in lines:
                if line.startswith("usemtl "):
                    material = line.strip().split()[-1]

                if line.startswith("f "):
                    fv, fn, fu = self.unpack_face(
                        line.split()[1:], vertices_all, normals_all, uv_all
                    )
                    if fv is not None:
                        vertices[material].extend(fv)
                    if fn is not None:
                        normals[material].extend(fn)
                    if fu is not None:
                        uv[material].extend(fu)

            # Gather all the parameters to create the objects by material
            objects = []
            for k in vertices.keys():
                material = self.default_material
                if self.materials.has_material(k):
                    self.materials.set_material(k)
                    material = Material(
                        ambient=self.materials.ambient,
                        diffuse=self.materials.diffuse,
                        specular=self.materials.specular,
                        Ns=self.materials.Ns,
                        texture=self.materials.texture,
                        bump_map=self.materials.bump_map
                    )
                objects.append({
                    "vertices": np.asarray(vertices[k]),
                    "normals": np.asarray(normals[k]) if len(normals[k]) > 0 else None,
                    "uv": np.asarray(uv[k]) if len(uv[k]) > 0 else None,
                    "material": material
                })

            return objects

        finally:
            close_file(f, filepath)

    def unpack_face(self, face, vertices, normals, uv):
        def triangulate_face(face):
            triangles = []
            if len(face) == 3:
                triangles.extend(face)
            else:
                for i in range(2, len(face)):
                    triangles.extend([face[0], face[i-1], face[i]])
            return triangles

        def extract_vertex(face):
            return int(face.split("/")[0])-1

        def extract_normal(face):
            return int(face.split("/")[2])-1

        def extract_uv(face):
            return int(face.split("/")[1])-1

        faces = triangulate_face(face)
        try:
            face_vertices = np.array(list(map(extract_vertex, faces)))
            face_vertices = vertices[face_vertices].reshape(-1, 3)
        except IndexError:
            face_vertices = None

        try:
            face_normals = np.array(list(map(extract_normal, faces)))
            face_normals = normals[face_normals].reshape(-1, 3)
        except IndexError:
            face_normals = None

        try:
            face_uv = np.array(list(map(extract_uv, faces)))
            face_uv = uv[face_uv].reshape(-1, uv.shape[1])[:, :2]
        except IndexError:
            face_uv = np.zeros((len(face_vertices), 2))

        return face_vertices, face_normals, face_uv



import numpy as np

from . import Behaviour


class RotateModel(Behaviour):
    """Rotate all the models around an axis with a given speed (radians per
    tick).
    
    Arguments
    ---------
        axis: {'x', 'y', 'z'}
        speed: float, radians per tick
    """
    def __init__(self, axis='z', speed=np.pi/90):
        self._function = dict(
            x=lambda s, a: s.rotate_x(a),
            y=lambda s, a: s.rotate_y(a),
            z=lambda s, a: s.rotate_z(a)
        )[axis]
        self._speed = speed

    def behave(self, params):
        self._function(params.scene, self._speed)
        params.refresh = True


class LocalModelRotation(Behaviour):
    """Use a per renderable model matrix to rotate the models if available.
    
    Arguments
    ---------
        axis: array-like (3,) the axis around which to rotate
        speed: float, radians per tick
    """
    def __init__(self, axis=[0, 0, 1.], speed=np.pi/90):
        self._axis = axis
        self._speed = speed

    def behave(self, params):
        refresh = False
        for r in params.scene.renderables:
            if hasattr(r, "rotate_axis"):
                r.rotate_axis(self._axis, self._speed)
                refresh = True
        params.refresh = refresh


class RotateRenderables(Behaviour):
    """Rotate specific renderables in the scene.

    Arguments
    ---------
        renderables: List of renderables to be rotated
        axis: array-like (3,) the axis around which to rotate
        speed: float, radians per tick
    """
    def __init__(self, renderables, axis=[0, 0, 1.], speed=np.pi/90):
        self.renderables = renderables
        self._axis = axis
        self._speed = speed

    def behave(self, params):
        for r in self.renderables:
            r.rotate_axis(self._axis, self._speed)
        params.refresh = True


class _TrajectoryMovement(Behaviour):
    """Abstract class that implements adjusting a quantity based on a passed
    trajectory.

    Arguments
    ---------
        trajectory: The Trajectory object that provides values
        speed: percentage of completion per tick
    """
    def __init__(self, trajectory, speed=0.01):
        self._trajectory = trajectory
        self._speed = speed
        self._t = 0.0

    def behave(self, params):
        v = self._trajectory.get_value(self._t)
        self._t += self._speed
        self._adjust(params, v)

    def _adjust(self, params, v):
        raise NotImplementedError()


class RenderableOffset(_TrajectoryMovement):
    """Move a renderable by adjusting their offset property.

    Arguments
    ---------
        renderable: A renderable to be moved
        trajectory: The Trajectory object that provides values
        speed: percentage of completion per tick
    """
    def __init__(self, renderable, trajectory, speed=0.01):
        super().__init__(trajectory, speed)
        self.renderable = renderable

    def _adjust(self, params, v):
        self.renderable.offset = v
        params.refresh = True


class CameraTrajectory(_TrajectoryMovement):
    """Move the camera along a trajectory."""
    def _adjust(self, params, v):
        params.scene.camera_position = v
        params.refresh = True


class CameraTargetTrajectory(_TrajectoryMovement):
    """Move the camera target along a trajectory."""
    def _adjust(self, params, v):
        params.scene.camera_target = v
        params.refresh = True


class LightTrajectory(_TrajectoryMovement):
    """Move the light along a trajectory."""
    def _adjust(self, params, v):
        params.scene.light = v
        params.refresh = True



import numpy as np
from pyrr import matrix33, vector

from . import Behaviour
from .trajectory import Circle


class MouseRotate(Behaviour):
    """Rotate the camera based using the mouse when left button is pressed.

    We rotate the camera with the following convention. At any given point we
    consider a sphere with a center in target and radius equal to the distance
    to the camera. We move on that sphere based on the movements of the mouse.
    """
    def __init__(self):
        self._start = None
        self._origin = None
        self._camera_pos = None
        self._right = None
        self._up = None

    def behave(self, params):
        if params.mouse.left_pressed:
            if self._start is None:
                self._start = params.mouse.location
                self._origin = params.scene.camera_target
                self._camera_pos = params.scene.camera_position
                cam_dir = vector.normalize(self._camera_pos - self._origin)
                self._right = np.cross(params.scene.up_vector, cam_dir)
                self._up = np.cross(cam_dir, self._right)
            else:
                size = params.scene.size
                end = params.mouse.location
                deltaX = float(end[0] - self._start[0])/size[0]
                deltaY = float(end[1] - self._start[1])/size[1]

                Rx = matrix33.create_from_axis_rotation(
                    self._up,
                    deltaX*2*np.pi
                )
                Ry = matrix33.create_from_axis_rotation(
                    self._right,
                    deltaY*2*np.pi
                )
                R = Ry.dot(Rx)
                newpos = self._origin + R.dot(self._camera_pos - self._origin)
                newup = R.dot(self._up)

                params.scene.camera_position = newpos
                params.scene.up_vector = newup
                params.refresh = True
        else:
            self._start = None


class MouseZoom(Behaviour):
    """Zoom in/out with the mouse scroll wheel."""
    def __init__(self, delta=0.9):
        self._delta = delta

    def behave(self, params):
        rotations = params.mouse.wheel_rotation
        if rotations != 0:
            cam_position = params.scene.camera_position
            cam_target = params.scene.camera_target
            ray = cam_target - cam_position
            if rotations > 0:
                cam_position += ray * (1-self._delta)
            else:
                cam_position -= ray * (1-self._delta)
            params.scene.camera_position = cam_position
            params.refresh = True


class MousePan(Behaviour):
    """Move the target by dragging the mouse with the middle button pressed."""
    def __init__(self, delta=1.):
        self._delta = delta
        self._start = None
        self._target = None
        self._right = None
        self._up = None

    def behave(self, params):
        if params.mouse.middle_pressed:
            if self._start is None:
                self._start = params.mouse.location
                self._target = params.scene.camera_target
                cam_dir = params.scene.camera_position - self._target
                cam_dir = vector.normalize(cam_dir)
                self._right = np.cross(params.scene.up_vector, cam_dir)
                self._up = np.cross(cam_dir, self._right)
            else:
                size = params.scene.size
                end = params.mouse.location
                deltaX = float(end[0] - self._start[0])/size[0]
                deltaY = float(end[1] - self._start[1])/size[1]

                newtarget = (
                    self._target +
                    -self._delta * deltaX * self._right +
                    self._delta * deltaY * self._up
                )
                params.scene.camera_target = newtarget
                params.refresh = True
        else:
            self._start = None



"""A scene defines the type of framebuffer to create and allows manipulation of
the programs uniforms, such as the camera position and light."""


import moderngl
import numpy as np
from pyrr import Matrix44, Vector3


class BaseScene(object):
    """Base scene holds the opengl context and defines utilities that will be
    useful for all scenes.
    
    Arguments
    ---------
        size: (width, height) of the framebuffer
        background: The color of the background frame buffer
        ctx: A moderngl context or None to create a standalone context
    """
    def __init__(self, size=(256, 256), background=(1, 1, 1, 1), ctx=None):
        if ctx is None:
            self._ctx = moderngl.create_standalone_context()
            self._ctx.enable(moderngl.BLEND)
            self._ctx.blend_func = (
                moderngl.SRC_ALPHA,
                moderngl.ONE_MINUS_SRC_ALPHA
            )
            self._framebuffer = self._ctx.framebuffer(
                self._ctx.renderbuffer(size),
                self._ctx.depth_renderbuffer(size)
            )
        else:
            self._ctx = ctx
            self._framebuffer = None

        self._background = background
        self._renderables = dict()

    @property
    def size(self):
        return self._ctx.fbo.size

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, v):
        b = list(self._background)
        for i in range(len(v)):
            b[i] = v[i]
        self._background = tuple(b)

    @property
    def renderables(self):
        return self._renderables.values()

    def add(self, renderable):
        if renderable not in self._renderables:
            renderable.init(self._ctx)
            self._renderables[renderable] = renderable

    def remove(self, renderable):
        if renderable in self._renderables:
            self._renderables[renderable].release()
            del self._renderables[renderable]

    def clear(self):
        for r in self._renderables:
            r.release()
        self._renderables.clear()

    @property
    def uniforms(self):
        """Return list of tuples with uniform name and value provided by the
        scene."""
        return []

    def render(self):
        # Update the uniforms
        uniforms = self.uniforms
        for r in self.renderables:
            r.update_uniforms(uniforms)

        # Render the scene
        if self._framebuffer is not None:
            self._framebuffer.use()
        self._ctx.enable(moderngl.DEPTH_TEST)
        self._ctx.clear(*self._background)
        for r in self.renderables:
            r.render()

    @property
    def frame(self):
        if self._framebuffer is not None:
            # NOTE: We flip the y axis in order to have (0, 0) to be the bottom
            # left corner.
            return np.frombuffer(
                self._framebuffer.read(components=4),
                dtype=np.uint8
            ).reshape(*(self._framebuffer.size + (4,)))[::-1]
        else:
            raise RuntimeError(("This scene is rendering on a framebuffer "
                                "managed by someone else (gui perhaphs). "
                                "Access the frame from there."))


class Scene(BaseScene):
    def __init__(self, size=(256, 256), background=(1, 1, 1, 1), ctx=None):
        super(Scene, self).__init__(size, background, ctx)

        self._camera = (
            Matrix44.perspective_projection(45., 1., 0.1, 1000.)
        ).astype(np.float32)
        self._camera_position = Vector3([-2., -2, -2], dtype="float32")
        self._camera_target = Vector3([0., 0, 0], dtype="float32")
        self._rotation = Matrix44.identity().astype(np.float32)
        self._up_vector = Vector3([0, 0, 1], dtype="float32")
        self._light = Vector3([-0.5, -0.8, -2], dtype="float32")
        self._mv = None

        self._uniforms = dict(
            # raw uniforms
            light=self._light,
            camera=self._camera,
            camera_position=self._camera_position,
            camera_target=self._camera_target,
            rotation=self._rotation,
            up_vector=self._up_vector,

            # derivative uniforms
            vm=self.vm,
            mv=self.mv,
            mvp=self.mvp
        )

    def _update_uniforms(self):
        self._uniforms["mvp"] = self.mvp
        self._uniforms["mv"] = self.mv
        self._uniforms["vm"] = self.vm

    @property
    def vm(self):
        return self.mv.inverse

    @vm.setter
    def vm(self, _vm):
        if _vm is None:
            self.mv = _vm
        else:
            _vm = Matrix44(_vm)
            self.mv = _vm.inverse

    @property
    def mv(self):
        if self._mv is None:
            return (Matrix44.look_at(
                self._camera_position,
                self._camera_target,
                self._up_vector
            ) * self._rotation).astype(np.float32)
        else:
            return self._mv.copy()

    @mv.setter
    def mv(self, _mv):
        if _mv is None:
            self._mv = _mv
        else:
            assert _mv.shape == (4, 4)
            self._mv = Matrix44(_mv.astype(np.float32).copy())

        self._update_uniforms()

    @property
    def mvp(self):
        return (self._camera * self.mv).astype(np.float32)

    @property
    def light(self):
        return self._light.copy()

    @light.setter
    def light(self, l):
        self._light[...] = l
        self._update_uniforms()

    @property
    def camera_matrix(self):
        return self._camera.copy()

    @camera_matrix.setter
    def camera_matrix(self, cam):
        self._camera[...] = cam
        self._update_uniforms()

    @property
    def camera_position(self):
        return self._camera_position.copy()

    @camera_position.setter
    def camera_position(self, pos):
        self._camera_position[...] = pos
        self._update_uniforms()

    @property
    def camera_target(self):
        return self._camera_target.copy()

    @camera_target.setter
    def camera_target(self, target):
        self._camera_target[...] = target
        self._update_uniforms()

    @property
    def up_vector(self):
        return self._up_vector.copy()

    @up_vector.setter
    def up_vector(self, up):
        self._up_vector[...] = up
        self._update_uniforms()

    @property
    def rotation(self):
        return self._rotation.copy()

    @rotation.setter
    def rotation(self, rot):
        self._rotation[...] = rot
        self._update_uniforms()

    def rotate_x(self, angle):
        self._rotation *= Matrix44.from_x_rotation(angle)
        self._update_uniforms()

    def rotate_y(self, angle):
        self._rotation *= Matrix44.from_y_rotation(angle)
        self._update_uniforms()

    def rotate_z(self, angle):
        self._rotation *= Matrix44.from_z_rotation(angle)
        self._update_uniforms()

    @property
    def uniforms(self):
        return list(self._uniforms.items())




import numpy as np
from pyrr import Matrix44, matrix44

from ..io import read_mesh_file
from ..io.voxels import read_binvox
from ..utils import normalize_colors
from .base import Renderable


class MeshBase(Renderable):
    """Abstract base class that implements functions commonly used by the
    subclasses."""
    def __init__(self, vertices, normals, offset=[0, 0, 0.]):
        self._vertices = np.asarray(vertices)
        self._normals = np.asarray(normals)
        self._model_matrix = np.eye(4).astype(np.float32)
        self._offset = np.asarray(offset).astype(np.float32)

        self._prog = None
        self._vbo = None
        self._vao = None

    def to_points_and_faces(self):
        N = 0
        index = {}
        faces = []
        for x in self._vertices.reshape(-1, 3):
            x = tuple(x)
            if x not in index:
                index[x] = N
                N += 1
            faces.append(index[x])
        points = [p for p, i in sorted(index.items(), key=lambda x: x[1])]

        return np.array(points), np.array(faces).reshape(-1, 3)

    @property
    def vertices(self):
        """Return all the vertices of this mesh. The vertices in groups of 3
        form triangles and several vertices occur multiple times as they are
        shared by several triangles."""
        return np.copy(self._vertices)

    @property
    def bbox(self):
        """The axis aligned bounding box of all the vertices as two
        3-dimensional arrays containing the minimum and maximum for each
        axis."""
        return [
            self._vertices.min(axis=0),
            self._vertices.max(axis=0)
        ]

    @property
    def model_matrix(self):
        """An affine transformation matrix (4x4) applied to the mesh before
        rendering. Can be changed to animate the mesh."""
        return self._model_matrix

    @model_matrix.setter
    def model_matrix(self, v):
        self._model_matrix = np.asarray(v).astype(np.float32)
        if self._prog:
            self._prog["local_model"].write(self._model_matrix.tobytes())

    def rotate_x(self, angle):
        """Helper function that multiplies the `model_matrix` with a rotation
        matrix around the x axis."""
        m = Matrix44.from_x_rotation(angle)
        self.model_matrix = m.dot(self.model_matrix)

    def rotate_y(self, angle):
        """Helper function that multiplies the `model_matrix` with a rotation
        matrix around the y axis."""
        m = Matrix44.from_y_rotation(angle)
        self.model_matrix = m.dot(self.model_matrix)

    def rotate_z(self, angle):
        """Helper function that multiplies the `model_matrix` with a rotation
        matrix around the z axis."""
        m = Matrix44.from_z_rotation(angle)
        self.model_matrix = m.dot(self.model_matrix)

    def rotate_axis(self, axis, angle):
        """Helper function that multiplies the `model_matrix` with a rotation
        matrix around the passed in axis."""
        m = matrix44.create_from_axis_rotation(axis, angle)
        self.model_matrix = m.dot(self.model_matrix)

    @property
    def offset(self):
        """A translation vector for the mesh vertices."""
        return self._offset

    @offset.setter
    def offset(self, v):
        self._offset = np.asarray(v).astype(np.float32)
        if self._prog:
            self._prog["offset"].write(self._offset.tobytes())

    def scale(self, s):
        """Multiply all the vertices with a number s."""
        self._vertices *= s
        self._update_vbo()

    def affine_transform(self, R=np.eye(3), t=np.zeros(3)):
        """Rotate and translate the vertices and then update the gpu buffer.

        Given the vertices v \in R^{Nx3} this function implements

            v' = v @ R + t

        Arguments
        ---------
            R: array (3, 3), the 3x3 rotation matrix
            t: array (3,), the translation vector
        """
        self._vertices = self._vertices.dot(R) + t
        self._update_vbo()

    def translate(self, t):
        """Translate all the vertices with a vector t."""
        self._vertices += t
        self._update_vbo()

    def to_unit_cube(self):
        """Transform the mesh such that it fits in the 0 centered unit cube."""
        bbox = self.bbox
        dims = bbox[1] - bbox[0]
        self._vertices -= dims/2 + bbox[0]
        self._vertices /= dims.max()
        self._update_vbo()

    def release(self):
        self._prog.release()
        self._vbo.release()
        self._vao.release()
        self._prog = None
        self._vbo = None
        self._vao = None

    def render(self):
        self._vao.render()

    def update_uniforms(self, uniforms):
        uniforms_list = self._get_uniforms_list()
        for k, v in uniforms:
            if k in uniforms_list:
                self._prog[k].write(v.tobytes())

    @staticmethod
    def _triangle_normals(triangles):
        triangles = triangles.reshape(-1, 3, 3)
        ba = triangles[:, 1] - triangles[:, 0]
        bc = triangles[:, 2] - triangles[:, 1]
        return np.cross(ba, bc, axis=-1)

    def _update_vbo(self):
        """Update the vertex buffer object because one of the values has
        changed (vertices, normals, etc)."""
        raise NotImplementedError()


class Mesh(MeshBase):
    """A mesh is a collection of triangles with normals and colors.

    Arguments
    ---------
        vertices: array-like, the vertices of the triangles. Each triangle
                  should be given on its own even if vertices are shared.
        normals: array-like, per vertex normal vectors
        colors: array-like, per vertex color as (r, g, b) floats or
                (r, g, b, a) floats. If one color is given then it is assumed
                to be for all vertices.
        offset: A translation vector for all the vertices. It can be changed
                after construction to animate the object together with the
                `model_matrix` property.
        mode: {"shading", "flat", "depth", "orthographic_depth"} A string that
              defines the rendering mode for this mesh. (default: shading)
        max_depth: The maximum depth map value when rendering depth maps
    """
    def __init__(self, vertices, normals, colors, offset=[0, 0, 0.],
                 mode="shading", max_depth=3.0):
        super(Mesh, self).__init__(vertices, normals, offset)

        self._colors = np.asarray(colors)

        N = len(self._vertices)
        if len(self._colors.shape) == 1:
            if self._colors.size == 3:
                self._colors = np.array(self._colors.tolist() + [1])
            self._colors = self._colors[np.newaxis].repeat(N, axis=0)
        elif self._colors.shape[1] == 3:
            self._colors = np.hstack([self._colors, np.ones((N, 1))])

        self._mode = mode
        self._max_depth = max_depth

    def init(self, ctx):
        self._prog = ctx.program(
            vertex_shader="""
                #version 330

                uniform mat4 mvp;
                uniform mat4 local_model;
                uniform vec3 offset;
                in vec3 in_vert;
                in vec3 in_norm;
                in vec4 in_color;
                out vec3 v_vert;
                out vec3 v_norm;
                out vec4 v_color;

                void main() {
                    v_color = in_color;

                    // Compute the position of the vertex
                    vec4 t_position = vec4(in_vert, 1.0);
                    t_position = local_model * t_position;
                    t_position = t_position + vec4(offset, 0.);
                    v_vert = t_position.xyz / t_position.w;
                    // TODO: We should probably use the global model matrix as
                    //       well but it is left for the future 
                    t_position = mvp * t_position;
                    gl_Position = t_position;

                    // Compute the normal of the vertex
                    vec4 t_normal = vec4(in_norm, 1.0);
                    t_normal = local_model * t_normal;
                    v_norm = t_normal.xyz / t_normal.w;
                }
            """,
            fragment_shader="""
                #version 330

                uniform vec3 light;
                uniform int mode;
                uniform float max_depth;
                uniform vec3 camera_position;
                in vec3 v_vert;
                in vec3 v_norm;
                in vec4 v_color;

                out vec4 f_color;

                void main() {
                    float lum = dot(normalize(v_norm), normalize(v_vert - light));
                    lum = acos(lum) / 3.14159265;
                    lum = clamp(lum, 0.0, 1.0);

                    // shading
                    if (mode == 0) {
                        f_color = vec4(v_color.xyz * lum, v_color.w);
                    }

                    // flat color
                    else if (mode == 1) {
                        f_color = v_color;
                    }

                    else if (mode == 2) {
                        float depth = distance(camera_position, v_vert) / max_depth;
                        depth = 1.0 - depth;
                        f_color = vec4(depth, depth, depth, 1);
                    }

                    // depth with orthographic projection
                    else {
                        float depth = 1-gl_FragCoord.z;
                        f_color = vec4(depth, depth, depth, 1);
                    }
                }
            """
        )
        self._vbo = ctx.buffer(np.hstack([
            self._vertices,
            self._normals,
            self._colors
        ]).astype(np.float32).tobytes())
        self._vao = ctx.simple_vertex_array(
            self._prog,
            self._vbo,
            "in_vert", "in_norm", "in_color"
        )
        self.model_matrix = self._model_matrix
        self.offset = self._offset
        self.mode = self._mode
        self.max_depth = self._max_depth

    @property
    def colors(self):
        """Return the color per vertex."""
        return self._colors.clone()

    @colors.setter
    def colors(self, c):
        c = normalize_colors(c, len(self._vertices))
        self._colors = c
        self._update_vbo()

    def _get_uniforms_list(self):
        """Return the used uniforms to fetch from the scene."""
        return ["light", "mvp", "camera_position"]

    def _update_vbo(self):
        """Write in the vertex buffer object the vertices, normals and
        colors."""
        if self._vbo is not None:
            self._vbo.write(np.hstack([
                self._vertices, self._normals, self._colors
            ]).astype(np.float32).tobytes())

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, new_mode):
        self._mode = new_mode
        modes = dict(
            shading=0,
            flat=1,
            depth=2,
            orthographic_depth=3
        )
        if self._prog:
            self._prog["mode"] = modes[self._mode]

    @property
    def max_depth(self):
        return self._max_depth

    @max_depth.setter
    def max_depth(self, new_max_depth):
        self._max_depth = new_max_depth
        if self._prog:
            self._prog["max_depth"] = self._max_depth

    def sort_triangles(self, point):
        """Sort the triangles such that the first is furthest from `point` and
        the last is the closest to `point`.

        It is used so that transparency works properly in OpenGL.
        """
        vertices = self._vertices.reshape(-1, 3, 3)
        normals = self._normals.reshape(-1, 9)
        colors = self._colors.reshape(-1, 12)

        centers = vertices.mean(-2)
        d = ((np.asarray(point).reshape(1, 3) - centers)**2).sum(-1)
        alpha = (colors[:, ::4].mean(-1)<1).astype(np.float32) * 1000
        idxs = np.argsort(d+alpha)[::-1]

        self._vertices = vertices[idxs].reshape(-1, 3)
        self._normals = normals[idxs].reshape(-1, 3)
        self._colors = colors[idxs].reshape(-1, 4)
        self._update_vbo()

    @classmethod
    def from_file(cls, filepath, color=(0.3, 0.3, 0.3), ext=None):
        """Read the mesh from a file.

        Arguments
        ---------
            filepath: Path to file or file object containing the mesh
            color: A default color to load if the information is not provided
                   in the file
            ext: The file extension (including the dot) if `filepath` is an
                 object
        """
        # Read the mesh
        mesh = read_mesh_file(filepath, ext=ext)

        # Extract the triangles
        vertices = mesh.vertices

        # Set a normal per triangle vertex
        try:
            normals = mesh.normals
        except NotImplementedError:
            normals = np.repeat(Mesh._triangle_normals(vertices), 3, axis=0)

        # Set a color per triangle vertex
        try:
            colors = mesh.colors
        except NotImplementedError:
            colors = np.ones((len(vertices), 1)) * color

        return cls(vertices, normals, colors)

    @classmethod
    def from_xyz(cls, X, Y, Z, colormap=None):
        X, Y, Z = list(map(np.asarray, [X, Y, Z]))
        def gray(x):
            return np.ones((x.shape[0], 3))*x[:, np.newaxis]

        def normalize(x):
            xmin = x.min()
            xmax = x.max()
            return 2*(x-xmin)/(xmax-xmin) - 1

        def idx(i, j, x):
            return i*x.shape[1] + j

        # Normalize dimensions in [-1, 1]
        x = normalize(X)
        y = normalize(Y)
        z = normalize(Z)

        # Create faces by triangulating each quad
        faces = []
        for i in range(x.shape[0]-1):
            for j in range(y.shape[1]-1):
                # i, j; i, j+1; i+1; j+1
                # i, j; i+1, j; i+1; j+1
                faces.extend([
                    idx(i+1, j+1, x),
                    idx(i, j+1, x),
                    idx(i, j, x),
                    idx(i+1, j, x),
                    idx(i+1, j+1, x),
                    idx(i, j, x)
                ])

        vertices = np.vstack([x.ravel(), y.ravel(), z.ravel()]).T[faces]
        colors = (
            colormap(Z.ravel()[faces])
            if colormap else gray(z.ravel()[faces])
        )
        normals = np.repeat(cls._triangle_normals(vertices), 3, axis=0)

        return cls(vertices, normals, colors)

    @classmethod
    def from_faces(cls, vertices, faces, colors=(0.3,0.3,0.3)):
        vertices, faces, colors = list(map(
            np.asarray,
            [vertices, faces, colors]
        ))
        colors = normalize_colors(colors, len(vertices))
        vertices = vertices[faces].reshape(-1, 3)
        normals = np.repeat(cls._triangle_normals(vertices), 3, axis=0)
        colors = colors[faces].reshape(-1, 4)

        return cls(vertices, normals, colors)

    @classmethod
    def from_boxes(cls, centers, sizes, colors=(0.3,0.3,0.3)):
        """Create boxes.

        Arguments
        ---------
            centers: Array of 3 dimensional centers
            sizes: Array of 3 sizes per box that give, half the width, half the
                   depth, half the height
            colors: tuple for all boxes or array of colors per box
        """
        box = np.array([[-1, -1,  1],
                        [ 1, -1,  1],
                        [ 1,  1,  1],
                        [-1, -1,  1],
                        [ 1,  1,  1],
                        [-1,  1,  1],
                        [-1,  1, -1],
                        [ 1,  1,  1],
                        [-1,  1,  1],
                        [-1,  1, -1],
                        [ 1,  1, -1],
                        [ 1,  1,  1],
                        [-1,  1, -1],
                        [-1, -1,  1],
                        [-1,  1,  1],
                        [-1,  1, -1],
                        [-1, -1,  1],
                        [-1, -1, -1],
                        [ 1, -1, -1],
                        [ 1, -1,  1],
                        [ 1,  1,  1],
                        [ 1, -1, -1],
                        [ 1,  1, -1],
                        [ 1,  1,  1],
                        [ 1, -1, -1],
                        [-1, -1,  1],
                        [ 1, -1,  1],
                        [ 1, -1, -1],
                        [-1, -1,  1],
                        [-1, -1, -1],
                        [ 1, -1, -1],
                        [-1,  1, -1],
                        [ 1,  1, -1],
                        [ 1, -1, -1],
                        [-1,  1, -1],
                        [-1, -1, -1]]).astype(np.float32)

        normals = np.array([[ 0,  0,  1],
                            [ 0,  0,  1],
                            [ 0,  0,  1],
                            [ 0,  0,  1],
                            [ 0,  0,  1],
                            [ 0,  0,  1],
                            [ 0,  1,  0],
                            [ 0,  1,  0],
                            [ 0,  1,  0],
                            [ 0,  1,  0],
                            [ 0,  1,  0],
                            [ 0,  1,  0],
                            [-1,  0,  0],
                            [-1,  0,  0],
                            [-1,  0,  0],
                            [-1,  0,  0],
                            [-1,  0,  0],
                            [-1,  0,  0],
                            [ 1,  0,  0],
                            [ 1,  0,  0],
                            [ 1,  0,  0],
                            [ 1,  0,  0],
                            [ 1,  0,  0],
                            [ 1,  0,  0],
                            [ 0, -1,  0],
                            [ 0, -1,  0],
                            [ 0, -1,  0],
                            [ 0, -1,  0],
                            [ 0, -1,  0],
                            [ 0, -1,  0],
                            [ 0,  0, -1],
                            [ 0,  0, -1],
                            [ 0,  0, -1],
                            [ 0,  0, -1],
                            [ 0,  0, -1],
                            [ 0,  0, -1]]).astype(np.float32)

        centers, sizes, colors = list(map(
            np.asarray,
            [centers, sizes, colors]
        ))

        assert len(centers.shape) == 2 and centers.shape[1] == 3
        if len(sizes.shape) == 1:
            sizes = sizes[np.newaxis].repeat(len(centers), axis=0)
        vertices = centers[:, np.newaxis]+sizes[:, np.newaxis]*box[np.newaxis]
        vertices = vertices.reshape(-1, 3)
        normals = np.vstack([normals]*len(centers))

        if len(colors.shape) == 1:
            if colors.size < 4:
                colors = np.array(colors.tolist() + [1.]*(4-colors.size))
            colors = colors[np.newaxis].repeat(len(vertices), axis=0)
        if len(colors) != len(vertices) and len(colors) == len(centers):
            colors = np.repeat(colors, len(box), axis=0)

        return cls(vertices, normals, colors)

    @classmethod
    def from_voxel_grid(cls, voxels, sizes=None, colors=(0.3, 0.3, 0.3),
                        bbox=[[-0.5, -0.5, -0.5], [0.5, 0.5, 0.5]]):
        """ Create a voxel grid

        Arguments
        ---------
            voxels: Array of 3D values, with true values indicating which
                    voxels to fill
            colors: The colors of the voxels. If colors is a vector then
                    it is the same for all voxels. If it is a 4 dimensional
                    tensor then a color per voxel is assumed.
        """
        # Make sure voxels, colors and bbox are arrays
        voxels, colors, bbox = list(map(np.asarray, [voxels, colors, bbox]))

        # Ensure that the voxel grid is indeed a 3D grid
        assert len(voxels.shape) == 3
        M, N, K = voxels.shape

        # Clean and standardize the sizes
        if sizes is None:
            sizes = (bbox[1]-bbox[0]) * 0.48 / [M, N, K]
        else:
            sizes = np.asarray(sizes)

        # Convert the indices to center coordinates
        x, y, z = np.indices((M, N, K)).astype(np.float32)
        x = x / M * (bbox[1][0] - bbox[0][0]) + bbox[0][0]
        y = y / N * (bbox[1][1] - bbox[0][1]) + bbox[0][1]
        z = z / K * (bbox[1][2] - bbox[0][2]) + bbox[0][2]
        centers = np.vstack([x[voxels], y[voxels], z[voxels]]).T

        # Convert the sizes to per box sizes
        if len(sizes.shape) == 1:
            sizes = np.array([sizes for _ in range(len(centers))])
        elif len(sizes.shape) == 4:
            sizes = sizes[voxels]

        # Convert the colors to per box colors
        if len(colors.shape) == 1:
            colors = np.array([colors for _ in range(len(centers))])
        elif len(colors.shape) == 4:
            colors = colors[voxels]

        return cls.from_boxes(centers=centers, sizes=sizes, colors=colors)

    @classmethod
    def from_binvox(cls, binvoxfile, colors=(0.3, 0.3, 0.3)):
        """Create a voxel grid from a binvox file.

        For the format see https://patrickmin.com/binvox/binvox.html .
        
        Arguments
        ---------
            binvoxfile: str or file object that contains the voxelgrid data in
                        binvox format
            colors: The colors of the voxels to pass to from_voxel_grid().
        """
        voxelgrid, translation, scale = read_binvox(binvoxfile)
        bbox = np.array([[0., 0, 0], [1, 1, 1]]) * scale + translation

        return cls.from_voxel_grid(voxelgrid, colors=colors, bbox=bbox)

    @classmethod
    def from_superquadrics(cls, alpha, epsilon, translation, rotation, colors,
                           offset=[0, 0, 0.], vertex_count=10000):
        """Create Superquadrics.

        Arguments
        ---------
            alpha: Array of 3 sizes, along each axis
            epsilon: Array of 2 shapes, along each a
            translation: Array of 3 dimensional center
            rotation: Array of size 3x3 containing the rotations
            colors: Tuple for all sqs or array of colors per sq
        """
        def fexp(x, p):
            return np.sign(x)*(np.abs(x)**p)

        def sq_surface(a1, a2, a3, e1, e2, eta, omega):
            x = a1 * fexp(np.cos(eta), e1) * fexp(np.cos(omega), e2)
            y = a2 * fexp(np.cos(eta), e1) * fexp(np.sin(omega), e2)
            z = a3 * fexp(np.sin(eta), e1)
            return x, y, z

        # triangulate the sphere to be used with the SQs
        n = int(np.sqrt(vertex_count))
        eta = np.linspace(-np.pi/2, np.pi/2, n, endpoint=True)
        omega = np.linspace(-np.pi, np.pi, n, endpoint=True)
        triangles = []
        for o1, o2 in zip(np.roll(omega, 1), omega):
            triangles.extend([
                (eta[0], 0),
                (eta[1], o2),
                (eta[1], o1),
            ])
        for e in range(1, len(eta)-2):
            for o1, o2 in zip(np.roll(omega, 1), omega):
                triangles.extend([
                    (eta[e], o1),
                    (eta[e+1], o2),
                    (eta[e+1], o1),
                    (eta[e], o1),
                    (eta[e], o2),
                    (eta[e+1], o2),
                ])
        for o1, o2 in zip(np.roll(omega, 1), omega):
            triangles.extend([
                (eta[-1], 0),
                (eta[-2], o1),
                (eta[-2], o2),
            ])
        triangles = np.array(triangles)
        eta, omega = triangles[:, 0], triangles[:, 1]

        # collect the pretriangulated vertices of each SQ
        vertices = []
        a, e, t, R = list(map(
            np.asarray,
            [alpha, epsilon, translation, rotation]
        ))
        M, _ = a.shape  # number of superquadrics
        assert R.shape == (M, 3, 3)
        assert t.shape == (M, 3)
        for i in range(M):
            a1, a2, a3 = a[i]
            e1, e2 = e[i]
            x, y, z = sq_surface(a1, a2, a3, e1, e2, eta, omega)
            # Get points on the surface of each SQ
            V = np.stack([x, y, z], axis=-1)
            V = R[i].T.dot(V.T).T + t[i].reshape(1, 3)
            vertices.append(V)

        # Finalize the mesh
        vertices = np.vstack(vertices)
        normals = np.repeat(cls._triangle_normals(vertices), 3, axis=0)
        colors = np.asarray(colors)

        if len(colors.shape) == 1:
            if colors.size < 4:
                colors = np.array(colors.tolist() + [1.]*(4-colors.size))
            colors = colors[np.newaxis].repeat(len(vertices), axis=0)
        assert len(colors) == len(vertices) or len(colors) == M
        if len(colors) == M:
            colors = np.repeat(colors, len(vertices) // M, axis=0)

        return cls(vertices, normals, colors, offset)



import moderngl
import numpy as np
import wx
import wx.glcanvas
from pyrr import Matrix44

from ..behaviours import Behaviour
from ..scenes import Scene
from .base import BaseWindow

class Window(BaseWindow):
    class _Frame(wx.Frame):
        """A simple frame for our wxWidgets app."""
        def __init__(self, window, size, title):
            super(Window._Frame, self).__init__(
                None,
                style=wx.DEFAULT_FRAME_STYLE
            )
            self._window = window
            self.SetTitle(title)
            self.SetClientSize(size)
            self.Center()
            self.view = Window._Canvas(self._window, self)
            self.Bind(wx.EVT_CLOSE, self._on_close)

        def _on_close(self, event):
            # If close was called before then close
            if self._window._closing:
                self.view._on_close(event)
                self.Destroy()

            # otherwise just set the window to closing
            self._window._closing = True

    class _Canvas(wx.glcanvas.GLCanvas):
        def __init__(self, window, parent):
            super(Window._Canvas, self).__init__(
                parent,
                attribList=[
                    wx.glcanvas.WX_GL_CORE_PROFILE,
                    wx.glcanvas.WX_GL_RGBA,
                    wx.glcanvas.WX_GL_DOUBLEBUFFER,
                    wx.glcanvas.WX_GL_DEPTH_SIZE,
                    24
                ]
            )
            self._window = window
            self._window._get_frame = self._get_frame
            self._context = wx.glcanvas.GLContext(self)
            self._mgl_context = None
            self._ticker = wx.Timer(self)

            self.Bind(wx.EVT_PAINT, self._on_paint)
            self.Bind(wx.EVT_TIMER, self._on_tick, self._ticker)
            self.Bind(wx.EVT_MOUSE_EVENTS, self._on_mouse)
            self.Bind(wx.EVT_KEY_DOWN, self._on_keyboard)
            self.Bind(wx.EVT_KEY_UP, self._on_keyboard)
            self.Bind(wx.EVT_SIZE, self._on_size)

        def _on_size(self, event):
            if self._mgl_context is not None:
                W, H = event.GetSize()
                self._mgl_context.viewport = (0, 0, W, H)
                self._window._scene.camera_matrix = (
                    Matrix44.perspective_projection(45., W/H, 0.1, 1000)
                )

        def _get_frame(self):
            W, H = self.GetClientSize()
            framebuffer = self._mgl_context.detect_framebuffer()
            return np.frombuffer(
                framebuffer.read(viewport=self._mgl_context.viewport, components=4),
                dtype=np.uint8
            ).reshape((H, W, 4))[::-1]

        def _on_close(self, event):
            self._ticker.Stop()
            self.Unbind(wx.EVT_TIMER)

        def _on_paint(self, event):
            self.SetCurrent(self._context)
            if self._window._scene is None:
                self._mgl_context = moderngl.create_context()
                self._mgl_context.enable(moderngl.BLEND)
                self._mgl_context.blend_func = (
                    moderngl.SRC_ALPHA,
                    moderngl.ONE_MINUS_SRC_ALPHA
                )
                self._window._scene = Scene(
                    size=(self.Size.width, self.Size.height),
                    background=(0,)*4,
                    ctx=self._mgl_context
                )
                self._ticker.Start(16)

            self._window._draw()
            self.SwapBuffers()

        def _on_tick(self, event):
            if self._window._behave(event):
                self.Refresh()
            if self._window._closing:
                self.GetParent()._on_close(None)
            self._window._mouse.wheel_rotation = 0
            self._window._keyboard.keys_up.clear()

        def _on_mouse(self, event):
            state = wx.GetMouseState()
            self._window._mouse.location = (state.GetX(), state.GetY())
            self._window._mouse.left_pressed = state.LeftIsDown()
            self._window._mouse.middle_pressed = state.MiddleIsDown()
            if abs(event.GetWheelRotation()) > 0:
                self._window._mouse.wheel_rotation += (
                    event.GetWheelRotation()/event.GetWheelDelta()
                )

        def _on_keyboard(self, event):
            down = event.GetEventType() == wx.EVT_KEY_DOWN.typeId
            key = chr(event.GetUnicodeKey())
            alt = event.AltDown() == down
            ctrl = event.ControlDown() == down
            cmd = event.CmdDown() == down
            meta = event.MetaDown() == down
            keys = set(
                [key] if key != "\00" else [] +
                (["<alt>"] if alt else []) +
                (["<ctrl>"] if ctrl else []) +
                (["<cmd>"] if cmd else []) +
                (["<meta>"] if meta else [])
            )

            if down:
                self._window._keyboard.keys_down.update(keys)
            else:
                self._window._keyboard.keys_up.update(
                    keys | self._window._keyboard.keys_down
                )
                self._window._keyboard.keys_down.difference_update(keys)

    def __init__(self, size=(512, 512)):
        super(Window, self).__init__(size)
        self._scene = None
        self._mouse = Behaviour.Mouse(None, None, None, None)
        self._keyboard = Behaviour.Keyboard([], [])
        self._closing = False

    def _behave(self, event):
        # Make the behaviour parameters
        params = Behaviour.Params(
            self,
            self._scene,
            self._get_frame,
            self._mouse,
            self._keyboard,
            self._closing
        )

        # Run the behaviours
        remove = []
        for i, b in enumerate(self._behaviours):
            b.behave(params)
            if params.done:
                remove.append(i)
                params.done = False
            if params.stop_propagation:
                break

        # Remove the ones that asked to be removed
        for i in reversed(remove):
            self._behaviours.pop(i)

        # If we are closing, remove all behaviours
        if self._closing:
            self._behaviours.clear()

        # Return whether we should paint again
        return params.refresh

    def _draw(self):
        self._scene.render()

    def show(self, title="Scene"):
        app = wx.App(False)
        frame = self._Frame(self, self.size, title)
        frame.Show()
        app.MainLoop()



import numpy as np

from ..utils import read_image


class Material(object):
    """A struct object containing information about the material.

    The supported materials have the following:

    - An ambient color
    - A diffuse lighting color (similar to `simple_3dviz.renderables.mesh.Mesh`)
    - A specular lighting color
    - A specular exponent for Phong lighting
    - A texture map
    - A bump map
    - A lighting mode from the set {'constant', 'diffuse', 'specular'}

    Arguments
    ---------
        ambient: array-like (r, g, b), float values between 0 and 1
        diffuse: array-like (r, g, b), float values between 0 and 1
        specular: array-like (r, g, b), float values between 0 and 1
        Ns: float, the exponent used for Phong lighting
        texture: array of uint8 with 3 or 4 channels and power of 2 width and
                 height, it contains the colors to be used by a mesh
        bump_map: array of uint8 with 3 channels and power of 2 width and
                  height, it contains the local displacement of the normal
                  vectors for implementing bump mapping
    """
    def __init__(self, ambient=(1.0, 1.0, 1.0), diffuse=(1.0, 1.0, 1.0),
                 specular=(0.1, 0.1, 0.1), Ns=2., texture=None,
                 bump_map=None, mode="diffuse"):
        self.ambient = np.asarray(ambient, dtype=np.float32)
        self.diffuse = np.asarray(diffuse, dtype=np.float32)
        self.specular = np.asarray(specular, dtype=np.float32)
        self.Ns = Ns
        self.texture = texture
        self.bump_map = bump_map
        if mode == "constant":
            self.diffuse[...] = 0
            self.specular[...] = 0
        elif mode == "diffuse":
            self.specular[...] = 0

    @property
    def texture_flipped(self):
        return self.texture[::-1]

    @property
    def bump_map_flipped(self):
        return self.bump_map[::-1]

    @classmethod
    def with_texture_image(cls, texture_path, ambient=(0.4, 0.4, 0.4),
                           diffuse=(0.4, 0.4, 0.4), specular=(0.1, 0.1, 0.1),
                           Ns=2., mode="specular"):
        return cls(
            ambient=ambient,
            diffuse=diffuse,
            specular=specular,
            Ns=Ns,
            texture=read_image(texture_path),
            mode=mode
        )




import numpy as np

from .utils import get_file, close_file


def read_binvox(filename):
    try:
        f = get_file(filename, "rb")

        # Read and validate the version line
        l = f.readline()
        assert l == b"#binvox 1\n"

        # Read the size of the voxel grid
        l = f.readline()
        assert l.startswith(b"dim")
        W, H, D = [int(d) for d in l.split()[1:]]

        # Read the translation vector
        l = f.readline()
        assert l.startswith(b"translate")
        translation = np.array([float(x) for x in l.split()[1:]])

        # Read the scaling factor
        l = f.readline()
        assert l.startswith(b"scale")
        scale = float(l.split()[1])

        # Make sure that the last line is 'data'
        assert f.readline().strip() == b"data"

        data = np.fromfile(f, dtype=np.uint8)
        values, counts = data[::2], data[1::2]
        voxelgrid = np.repeat(values.astype(np.bool), counts)
        voxelgrid = voxelgrid.reshape(W, H, D).transpose(0, 2, 1)

        return voxelgrid, translation, scale

    finally:
        close_file(filename, f)




from ..utils import save_frame
from . import Behaviour


class OnKeys(Behaviour):
    """Abstract class that performs an action when a set of keys is pressed."""
    def __init__(self, keys):
        self._keys = keys
        self._act = False

    def behave(self, params):
        if params.keyboard.keys_down & self._keys == self._keys:
            self._act = True
        elif self._act and params.keyboard.keys_up | self._keys:
            self._act = False
            self.action(params)

    def action(self, params):
        raise NotImplementedError()


class SnapshotOnKey(OnKeys):
    """Save a snapshot when a key is pressed.
    
    Arguments
    ---------
        path: The path to save the images with the optional index for the
              snapshot (ie. snapshot_{}.png)
        keys: A set of keys that must all be pressed to save the frame.
              Available special keys are <ctrl>, <cmd>, <alt>, etc.
    """
    def __init__(self, path="snapshot_{:03d}.png", keys={"S"}):
        super(SnapshotOnKey, self).__init__(keys)
        self._path = path
        self._i = 0

    def action(self, params):
        frame = params.frame()
        path = self._path.format(self._i)
        self._i += 1
        save_frame(path, frame)


class CameraReport(OnKeys):
    """Print the scene parameters to stdout when a key is pressed.

    Arguments
    ---------
        keys: A set of keys that must all be pressed to activate the action.
    """
    def __init__(self, keys={"R"}):
        super(CameraReport, self).__init__(keys)

    def action(self, params):
        def csv(x):
            return ",".join(list(map(str, x)))
        print("--camera_target='{}' --camera_position='{}' --up='{}'".format(
            csv(params.scene.camera_target),
            csv(params.scene.camera_position),
            csv(params.scene.up_vector)
        ))


class SortTriangles(OnKeys):
    """Sort the triangles wrt to the camera so that the transparency works.

    Arguments
    ---------
        keys: A set of keys that must all be pressed to activate the action
    """
    def __init__(self, keys={"T"}):
        super(SortTriangles, self).__init__(keys)

    def action(self, params):
        camera = params.scene.camera_position
        for r in params.scene.renderables:
            if hasattr(r, "sort_triangles"):
                r.sort_triangles(camera)
        params.refresh = True



"""The scripts module contains executable modules that provide simple utility
scripts showcasing the abilities of simple_3dviz and allowing for the quick
visualization of two-dimensional functions (`simple_3dviz.scripts.func_viewer`)
and meshes (`simple_3dviz.scripts.mesh_viewer`)."""



from os import path

import numpy as np

from ..io import read_mesh_file, read_material_file
from ..io.multi_mesh import MultiMaterialObjReader
from .base import RenderableCollection
from .mesh import MeshBase
from .material import Material


class TexturedMesh(MeshBase):
    """A mesh that can use materials and textures to render a slightly more
    realistic image.

    Arguments
    ---------
        vertices: array-like, the vertices of the triangles. Each triangle
                  should be given on its own even if vertices are shared.
        normals: array-like, per vertex normal vectors
        uv: array-like, per-vertex uv coordinates inside the texture
        material: simple_3dviz.renderables.textured_mesh.Material object
    """
    def __init__(self, vertices, normals, uv, material):
        vertices = np.asarray(vertices)

        # If the normals are not provided compute them from the faces
        if normals is None:
            normals = np.repeat(self._triangle_normals(vertices), 3, axis=0)

        super(TexturedMesh, self).__init__(vertices, normals)

        # If the uv coordinates are not provided set them to 0
        if uv is None:
            uv = np.zeros((vertices.shape[0], 2), dtype=np.float32)

        self._uv = np.asarray(uv)
        assert(self._uv.shape == (len(self._vertices), 2))
        self._material = material

        self._texture = None
        self._bump_map = None
        self._cull_back_face = True

    def init(self, ctx):
        self._prog = ctx.program(
            vertex_shader="""
                #version 330

                uniform mat4 mvp;
                uniform mat4 rotation;
                uniform mat4 local_model;
                uniform vec3 offset;
                in vec3 in_vert;
                in vec3 in_norm;
                in vec2 in_uv;
                out vec3 v_vert;
                out vec3 v_norm;
                out vec2 v_uv;

                void main() {
                    vec4 t_pos = vec4(in_vert, 1.0);
                    vec3 t_nor = in_norm;

                    t_pos = local_model * t_pos;
                    t_pos = t_pos + vec4(offset, 0);
                    vec3 g_pos = vec3(t_pos);
                    t_nor = mat3(local_model) * t_nor;
                    t_nor = mat3(rotation) * t_nor;
                    t_pos = mvp * t_pos;

                    // outputs
                    v_vert = g_pos;
                    v_norm = t_nor;
                    v_uv = in_uv;
                    gl_Position = t_pos;
                }
            """,
            fragment_shader="""
                #version 330

                uniform vec3 light;
                uniform vec3 camera_position;
                uniform vec3 ambient;
                uniform vec3 diffuse;
                uniform vec3 specular;
                uniform float Ns;
                uniform sampler2D texture;
                uniform sampler2D bump_map;
                uniform bool has_texture;
                uniform bool has_bump_map;
                uniform bool cull_back_face;
                in vec3 v_vert;
                in vec3 v_norm;
                in vec2 v_uv;

                out vec4 f_color;

                void main() {
                    // Local variables for the colors and normal
                    vec3 l_ambient = ambient;
                    vec3 l_diffuse = diffuse;
                    vec3 l_norm = v_norm;

                    // Discard pixels where the normal is pointing away from the camera.
                    if (cull_back_face) {
                    vec3 T = normalize(camera_position - v_vert);
                        if (dot(T, l_norm) < 0) {
                            discard;
                            return;
                        }
                    }

                    // fix colors based on the textures
                    if (has_texture) {
                        vec4 texColor = texture2D(texture, v_uv);
                        l_ambient = l_ambient * texColor.rgb;
                        l_diffuse = l_diffuse * texColor.rgb;
                    }

                    // fix normal based on the bump map
                    if (has_bump_map) {
                        vec3 bump_normal = texture2D(bump_map, v_uv).rgb;
                        l_norm += (dot(v_norm, v_norm) - 1) * bump_normal;
                    }

                    // ambient color
                    f_color.rgb = l_ambient;

                    // diffuse color
                    vec3 L = normalize(light - v_vert);
                    // The theory says the commented line however the one used looks better
                    // f_color.rgb += l_diffuse * clamp(dot(normalize(l_norm), L), 0, 1);
                    f_color.rgb += l_diffuse * clamp(abs(dot(normalize(l_norm), L)), 0.5, 1) * 1.4;

                    // specular color
                    if (Ns > 0) {
                        vec3 V = normalize(camera_position - v_vert);
                        vec3 H = normalize(L + V);
                        f_color.rgb += specular * pow(clamp(dot(normalize(l_norm), H), -1, 1), Ns);
                        f_color.a = 1.0;
                    }
                }
            """
        )
        self._vbo = ctx.buffer(np.hstack([
            self._vertices,
            self._normals,
            self._uv
        ]).astype(np.float32).tobytes())
        self._vao = ctx.simple_vertex_array(
            self._prog,
            self._vbo,
            "in_vert", "in_norm", "in_uv"
        )
        self._prog["texture"] = 0
        self._prog["bump_map"] = 1
        self.model_matrix = self._model_matrix
        self.offset = self._offset
        self.material = self._material
        self.cull_back_face = self._cull_back_face

    def release(self):
        super(TexturedMesh, self).release()
        if self._texture is not None:
            self._texture.release()
            self._texture = None
        if self._bump_map is not None:
            self._bump_map.release()
            self._bump_map = None

    def _get_uniforms_list(self):
        """Return the used uniforms to fetch from the scene."""
        return ["light", "camera_position", "mvp", "rotation"]

    def _update_vbo(self):
        """Write in the vertex buffer object the vertices, normals and
        colors."""
        if self._vbo is not None:
            self._vbo.write(np.hstack([
                self._vertices, self._normals, self._uv, self._material
            ]).astype(np.float32).tobytes())

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, new_material):
        assert isinstance(new_material, Material)
        self._material = new_material
        if self._prog is None:
            return

        self._prog["ambient"].write(self._material.ambient.tobytes())
        self._prog["diffuse"].write(self._material.diffuse.tobytes())
        self._prog["specular"].write(self._material.specular.tobytes())
        self._prog["Ns"] = self._material.Ns
        if self._material.texture is not None:
            self._prog["has_texture"] = True
            if self._texture is not None:
                self._texture.release()
            self._texture = self._prog.ctx.texture(
                self._material.texture.shape[:2][::-1],
                self._material.texture.shape[2],
                data=self._material.texture_flipped.tobytes()
            )
        else:
            self._prog["has_texture"] = False
        if self._material.bump_map is not None:
            self._prog["has_bump_map"] = True
            if self._bump_map is not None:
                self._bump_map.release()
            self._bump_map = self._prog.ctx.texture(
                self._material.bump_map.shape[:2][::-1],
                self._material.bump_map.shape[2],
                data=self._material.bump_map_flipped.tobytes()
            )
        else:
            self._prog["has_bump_map"] = False

    @property
    def cull_back_face(self):
        """Completely discard triangles that are facing away from the
        camera."""
        return self._cull_back_face

    @cull_back_face.setter
    def cull_back_face(self, cull):
        self._cull_back_face = bool(cull)
        if self._prog:
            self._prog["cull_back_face"] = self._cull_back_face

    def render(self):
        if self._texture is not None:
            self._texture.use(location=0)
        if self._bump_map is not None:
            self._bump_map.use(location=1)
        super(TexturedMesh, self).render()

    @classmethod
    def from_file(cls, filepath, material_filepath=None, ext=None,
                  material_ext=None, color=(0.5, 0.5, 0.5)):
        """Read the mesh from a file.

        Arguments
        ---------
            filepath: Path to file or file object containing the mesh
            material_filepath: Path to file containing the material
            ext: The file extension (including the dot) if 'filepath' is an
                 object
            material_ext: The file extension (including the dot) if 'filepath'
                          is an object
            color: A color to use if the material is neither given nor found
        """
        if isinstance(filepath, str) and filepath.endswith(".obj") or ext == ".obj":
            return cls.from_obj_file(filepath, material_filepath, material_ext, color)

        # Read the mesh
        mesh = read_mesh_file(filepath, ext=ext)

        # Extract the triangles
        vertices = mesh.vertices

        # Set a normal per triangle vertex
        try:
            normals = mesh.normals
        except NotImplementedError:
            normals = np.repeat(cls._triangle_normals(vertices), 3, axis=0)

        # Set the uv coordinates
        try:
            uv = mesh.uv
        except NotImplementedError:
            uv = np.zeros((vertices.shape[0], 2), dtype=np.float32)

        # Parse the material information
        mtl = None
        material = Material(diffuse=color)
        try:
            mtl = read_material_file(mesh.material_file)
        except (NotImplementedError, FileNotFoundError):
            if material_filepath is not None:
                mtl = read_material_file(material_filepath, ext=material_ext)
            elif path.exists(path.join(path.dirname(filepath), "texture.png")):
                try:
                    material = Material.with_texture_image(
                        path.join(path.dirname(filepath), "texture.png"),
                        diffuse=color
                    )
                except:
                    import sys
                    print(("Error while reading the texture image "
                           "for {}").format(filepath), file=sys.stderr)
                    raise

        if mtl is not None:
            material = Material(
                ambient=mtl.ambient,
                diffuse=mtl.diffuse,
                specular=mtl.specular,
                Ns=mtl.Ns,
                texture=mtl.texture,
                bump_map=mtl.bump_map
           )

        return cls(vertices, normals, uv, material)

    @classmethod
    def from_faces(cls, vertices, uv, faces, material):
        vertices, uv, faces = map(np.asarray, (vertices, uv, faces))
        vertices = vertices[faces].reshape(-1, 3)
        uv = uv[faces].reshape(-1, 2)
        normals = np.repeat(cls._triangle_normals(vertices), 3, axis=0)

        return cls(vertices, normals, uv, material)


    @classmethod
    def from_obj_file(cls, filepath, material_filepath=None, material_ext=None,
                      color=(0.5, 0.5, 0.5)):
        reader = MultiMaterialObjReader(filepath, material_filepath, material_ext, color)
        renderables = [cls(**object_args) for object_args in reader.objects]
        if len(renderables) == 1:
            return renderables[0]
        else:
            return RenderableCollection(renderables)



"""This module defines and implements the Trajectory interface which provides a
way to generate a sequence of values for use with behaviours. For instance we
can generate quadratic bezier trajectories for moving the camera or lights."""

from bisect import bisect_right

import numpy as np
from pyrr import matrix33


class Trajectory(object):
    """The main trajectory interface takes a percentage and provides a
    value (most often 3d position but also color etc.)."""
    def get_value(self, t):
        """Given t in [0, 1] return the value of the trajectory. Some
        trajectories are periodic which means they support t > 1."""
        raise NotImplementedError()


class StartStopTrajectory(Trajectory):
    """Start and stop the decorated trajectory for the given values of t by
    mapping the interval to 0-1 and everything outside the interval to either 0
    or 1."""
    def __init__(self, trajectory, start=0.0, stop=1.0):
        self._trajectory = trajectory
        self._start = start
        self._stop = stop

    def get_value(self, t):
        tt = (t - self._start) / (self._stop - self._start)
        return self._trajectory.get_value(min(1.0, max(0.0, tt)))


class Linear(Trajectory):
    """From a to b the fastest way possible :-)."""
    def __init__(self, a, b):
        self._a = a
        self._b = b

    def get_value(self, t):
        if not (0 <= t <= 1):
            raise ValueError(("Linear trajectory is defined only for "
                              "t in [0, 1]"))

        return self._a + t * (self._b - self._a)


class Join(Trajectory):
    """Join several trajectories into one large trajectory using a fixed
    percentage for each.

    The passed in trajectories should be tuples of weight, trajectory. For
    instance:

        t = Join([
            (0.1, Linear(0, 1)),
            (0.5, Linear(1, 0)),
            (0.1, Linear(0, -1)),
            (0.5, Linear(-1, 0))
        ])

    NOTE: The weights don't have to sum to 1.
    """
    def __init__(self, trajectories):
        self._weights = [t[0] for t in trajectories]
        for i in range(1, len(self._weights)):
            self._weights[i] = self._weights[i] + self._weights[i-1]
        self._trajectories = [t[1] for t in trajectories]
        if not all(isinstance(t, Trajectory) for t in self._trajectories):
            raise ValueError(("The trajectories passed should implement the "
                              "Trajectory interface"))

    def get_value(self, t):
        if not (0 <= t <= 1):
            raise ValueError(("Join trajectory is defined only for "
                              "t in [0, 1]"))

        # Find the correct trajectory
        w = t*self._weights[-1]
        index = bisect_right(self._weights, w)
        if index == len(self._trajectories):
            return self._trajectories[-1].get_value(1.)

        # Find the completion percentage of that trajectory
        max_weight = self._weights[index]
        prev_weight = 0. if index==0 else self._weights[index-1]
        new_t = (w-prev_weight)/(max_weight-prev_weight)

        return self._trajectories[index].get_value(new_t)


class Repeat(Trajectory):
    """Repeat a trajectory assuming that the end point is the same as the start
    point.

    For instance can be used with Lines as follows:

        Repeat(Lines([0, 0, 0], [0, 1, 0], [1, 1, 0], [1, 0, 0], [0, 0, 0]))
    """
    def __init__(self, trajectory):
        self._trajectory = trajectory

    def get_value(self, t):
        return self._trajectory.get_value(t % 1)


class BackAndForth(Trajectory):
    """Run a trajectory forwards and backwards continuously."""
    def __init__(self, trajectory):
        self._trajectory = trajectory

    def get_value(self, t):
        t = t % 2
        return self._trajectory.get_value(2-t if t > 1 else t)


class Circle(Trajectory):
    """Circle implements a clockwise circular trajectory in 3d space.

    The circle is defined by the center, a 3d point and a normal vector. The
    start of the circle is the 3d point that rotates around the normal
    vector.
    """
    def __init__(self, center, point, normal):
        self._center = np.asarray(center, dtype=np.float64)
        self._point = np.asarray(point, dtype=np.float64)
        self._normal = np.asarray(normal, dtype=np.float64)
        radial = self._point - self._center

        cosangle = (
            self._normal.dot(radial) /
            np.sqrt(self._normal.dot(self._normal)) /
            np.sqrt(radial.dot(radial))
        )
        if cosangle > 0.01:
            raise ValueError(("The radial vector and the normal vector "
                              "are not perpendicular thus do not define "
                              "a circle"))

    def get_value(self, t):
        R = matrix33.create_from_axis_rotation(self._normal, 2*t*np.pi)
        return self._center + R.dot(self._point-self._center)


class QuadraticBezier(Trajectory):
    """A simple quadratic bezier curve.

    https://en.wikipedia.org/wiki/B%C3%A9zier_curve
    """
    def __init__(self, A, B, C):
        self._A = A
        self._B = B
        self._C = C

    def get_value(self, t):
        if not (0 <= t <= 1):
            raise ValueError(("QuadraticBezier trajectory is defined only for "
                              "t in [0, 1]"))

        return (1-t)**2 * self._A + 2*(1-t)*t * self._B + t**2 * self._C


def Lines(*points):
    """Lines is a helper function to create joins of linear trajectories.

    Lines accepts points as positional arguments or as a single iterable. The
    arguments are cast to numpy arrays and then a trajectory is created.
    """
    if len(points) > 1:
        points = np.asarray(points, dtype=np.float64)
    else:
        points = np.asarray(points[0], dtype=np.float64)

    assert len(points) > 1, "Lines expects at least two points"

    return Join([
        (1., Linear(points[i], points[i+1]))
        for i in range(len(points)-1)
    ])


def QuadraticBezierCurves(*points):
    """QuadraticBezierCurves is a helper function to create joins of quadratic
    bezier curves.

    Same as Lines it accepts points either as positional arguments or a single
    iterable.
    """
    if len(points) > 1:
        points = np.asarray(points, dtype=np.float64)
    else:
        points = np.asarray(points[0], dtype=np.float64)

    assert len(points) > 2, "QuadraticBezierCurves expects at least 3 points"
    assert (len(points) % 2) == 1, ("QuadraticBezierCurves expects an odd "
                                    "number of points")

    return Join([
        (1., QuadraticBezier(points[i], points[i+1], points[i+2]))
        for i in range(0, len(points)-2, 2)
    ])




import moderngl
import numpy as np

from ..io.voxels import read_binvox
from .base import Renderable


class Lines(Renderable):
    """A line is a collection of line segments with colors and a specific width.

    Arguments:
    ----------
        points: array-like, the points that compose the line segments.
        colors: array-like, per line-segment color as (r,g,b,a)
        width: float indicating the width of the line
    """
    def __init__(self, points, colors=(0.3, 0.3, 0.3, 1.0), width=0.4):
        self._points = np.asarray(points)
        self._colors = np.asarray(colors)
        self._width = width

        N = len(self._points)
        if len(self._colors.shape) == 1:
            if self._colors.size == 3:
                self._colors = np.array(self._colors.tolist() + [1])
            self._colors = self._colors[np.newaxis].repeat(N, axis=0)
        elif self._colors.shape[1] == 3:
            self._colors = np.hstack([self._colors, np.ones((N, 1))])

        self._prog = None
        self._vbo = None
        self._vao = None

    def init(self, ctx):
        self._prog = ctx.program(
            vertex_shader="""
                #version 330

                in vec3 in_vertex;
                in vec4 in_color;
                out vec4 v_color;

                void main() {
                    v_color = in_color;
                    gl_Position = vec4(in_vertex, 1);
                }
            """,
            geometry_shader="""
                #version 330

                layout(lines) in;
                layout(triangle_strip, max_vertices=4) out;

                uniform float width;
                uniform mat4 vm;
                uniform mat4 mvp;
                in vec4 v_color[];
                out vec4 t_color;

                void main() {
                    vec3 camera_position = vm[3].xyz / vm[3].w;
                    vec3 first_v = gl_in[0].gl_Position.xyz;
                    vec3 last_v = gl_in[1].gl_Position.xyz;

                    vec3 ray_first = normalize(camera_position - first_v);
                    vec3 ray_last = normalize(camera_position - last_v);
                    vec3 line = normalize(last_v - first_v);
                    vec3 offset_first = cross(ray_first, line)*width/2;
                    vec3 offset_last = cross(ray_last, line)*width/2;

                    gl_Position = mvp * vec4(first_v + offset_first, 1);
                    t_color = v_color[0];
                    EmitVertex();
                    gl_Position = mvp * vec4(first_v - offset_first, 1);
                    t_color = v_color[0];
                    EmitVertex();
                    gl_Position = mvp * vec4(last_v + offset_last, 1);
                    t_color = v_color[1];
                    EmitVertex();
                    gl_Position = mvp * vec4(last_v - offset_last, 1);
                    t_color = v_color[1];
                    EmitVertex();

                    EndPrimitive();
                }
            """,
            fragment_shader="""
                #version 330

                in vec4 t_color;
                out vec4 f_color;

                void main() {
                    f_color = t_color;
                }
            """
        )
        self._vbo = ctx.buffer(
            np.hstack([self._points, self._colors]).astype(np.float32).tobytes()
        )
        self._vao = ctx.simple_vertex_array(
            self._prog,
            self._vbo,
            "in_vertex", "in_color"
        )

    def release(self):
        self._prog.release()
        self._vbo.release()
        self._vao.release()

    def render(self):
        self._vao.render(moderngl.LINES)

    def update_uniforms(self, uniforms):
        for k, v in uniforms:
            if k in ["mvp", "vm"]:
                self._prog[k].write(v.tobytes())
        self._prog["width"].value = self._width

    @property
    def bbox(self):
        """The axis aligned bounding box of all the vertices as two
        3-dimensional arrays containing the minimum and maximum for each
        axis."""
        return [
            self._points.min(axis=0),
            self._points.max(axis=0)
        ]

    def to_unit_cube(self):
        bbox = self.bbox
        dims = bbox[1] - bbox[0]
        self._points -= dims/2 + bbox[0]
        self._points /= dims.max()
        if self._vbo is not None:
            self._vbo.write(np.hstack([
                self._points, self._colors
            ]).astype(np.float32).tobytes())

    @classmethod
    def from_voxel_grid(cls, voxels, colors=(0.1, 0.1, 0.1), width=0.001,
                        bbox=[[-0.5, -0.5, -0.5], [0.5, 0.5, 0.5]]):
        """Create a voxel grid wire frame.

        Arguments
        ---------
            voxels: Array of 3D values, with truthy values indicating which
                    voxels to fill
        """
        # Make sure voxels, colors and bbox are arrays
        voxels, colors, bbox = list(map(np.asarray, [voxels, colors, bbox]))

        # Ensure that the voxel grid is indeed a 3D grid
        assert len(voxels.shape) == 3
        M, N, K = voxels.shape

        # Compute the size of each side
        sizes = 0.5 * (bbox[1] - bbox[0]) / [M, N, K]

        # Convert the indices to center coordinates
        x, y, z = np.indices((M, N, K)).astype(np.float32)
        x = x / M * (bbox[1][0] - bbox[0][0]) + bbox[0][0]
        y = y / N * (bbox[1][1] - bbox[0][1]) + bbox[0][1]
        z = z / K * (bbox[1][2] - bbox[0][2]) + bbox[0][2]
        centers = np.vstack([x[voxels], y[voxels], z[voxels]]).T

        # Create an array containing the cube edges
        edges = np.array([[-1, -1, -1], [ 1, -1, -1],
                          [ 1, -1, -1], [ 1,  1, -1],
                          [ 1,  1, -1], [-1,  1, -1],
                          [-1,  1, -1], [-1, -1, -1],
                          [-1, -1,  1], [ 1, -1,  1],
                          [ 1, -1,  1], [ 1,  1,  1],
                          [ 1,  1,  1], [-1,  1,  1],
                          [-1,  1,  1], [-1, -1,  1],
                          [-1, -1, -1], [-1, -1,  1],
                          [-1,  1, -1], [-1,  1,  1],
                          [ 1, -1, -1], [ 1, -1,  1],
                          [ 1,  1, -1], [ 1,  1,  1]]) * sizes

        # Finally create the edges of each cube
        points = centers[:, np.newaxis] + edges[np.newaxis]

        # Convert the colors to per edge color
        if len(colors.shape) == 1:
            colors = np.array([colors]*(points.size // 3))
        elif len(colors.shape) == 4:
            colors = colors[voxels]
            colors = np.repeat(colors, len(edges), axis=0)

        return cls(points.reshape(-1, 3), colors, width)

    @classmethod
    def from_binvox(cls, binvoxfile, colors=(0.1, 0.1, 0.1), width=0.001):
        """Create a wireframe for voxel grid read from a binvox file.

        Arguments
        ---------
            binvoxfile: str or file object that contains the voxelgrid data in
                        binvox format
            colors: The colors of the voxels to pass to from_voxel_grid().
            width: The width of the lines for the wireframe.
        """
        voxelgrid, translation, scale = read_binvox(binvoxfile)
        bbox = np.array([[0., 0, 0], [1, 1, 1]]) * scale + translation

        return cls.from_voxel_grid(voxelgrid, colors=colors, bbox=bbox,
                                   width=width)

    @classmethod
    def axes(cls, origin=(0, 0, 0), size=1.0, colors=None, width=0.01):
        """Create the three axes to be used as a reference.

        Arguments
        ---------
            origin: array-like (3,), the origin to put the axes to
                    (default: (0, 0, 0))
            size: float or array-like (3,), the size of the axes lines
                  (default: 1.)
            colors: None or array-like (3, 3 or 4), the colors to use for each
                    of the three axes (default: None)
            width: float, the width of the lines (default: 0.01)
        """
        # Normalize the colors argument
        colors = colors or [[0.8, 0.2, 0.2], [0.2, 0.8, 0.2], [0.2, 0.2, 0.8]]
        colors = np.asarray(colors)
        if len(colors.shape) == 1:
            colors = colors[None]
        if len(colors) == 1:
            colors = np.repeat(colors, 3, axis=0)
        elif len(colors) != 3:
            raise ValueError("colors should contain 1 or 3 colors")
        if colors.shape[1] == 3:
            colors = np.hstack([colors, np.ones((3, 1))])
        elif colors.shape[1] != 4:
            raise ValueError("colors should be either 3 or 4 values")
        colors = np.repeat(colors, 2, axis=0)
        assert colors.shape == (6, 4)

        # Normalize the size argument
        size = np.ones(3) * size
        assert size.shape == (3,)

        # Normalize the origin argument
        origin = np.asarray(origin)
        assert origin.shape == (3,)

        axes = np.array([[0, 0, 0],
                         [1, 0, 0],
                         [0, 0, 0],
                         [0, 1, 0],
                         [0, 0, 0],
                         [0, 0, 1.]]) * size  + origin

        return cls(axes, colors, width=width)



#!/usr/bin/env python

"""Setup simple-3dviz"""

from itertools import dropwhile
import os
from os import path
from setuptools import find_packages, setup


def collect_docstring(lines):
    """Return document docstring if it exists"""
    lines = dropwhile(lambda x: not x.startswith('"""'), lines)
    doc = ""
    for line in lines:
        doc += line
        if doc.endswith('"""\n'):
            break

    return doc[3:-4].replace("\r", "").replace("\n", " ")


def collect_metadata():
    meta = {}
    with open(path.join("simple_3dviz", "__init__.py")) as f:
        lines = iter(f)
        meta["description"] = collect_docstring(lines)
        for line in lines:
            if line.startswith("__"):
                key, value = map(lambda x: x.strip(), line.split("="))
                meta[key[2:-2]] = value[1:-1]

    return meta


def get_extensions():
    return []


def get_install_requirements():
    return [
        "moderngl",
        "numpy",
        "pyrr",
        "plyfile",
        "Pillow"
    ]


def setup_package():
    with open("README.md") as f:
        long_description = f.read()
    meta = collect_metadata()
    setup(
        name="simple-3dviz",
        version=meta["version"],
        description=meta["description"],
        long_description=long_description,
        long_description_content_type="text/markdown",
        maintainer=meta["maintainer"],
        maintainer_email=meta["email"],
        url=meta["url"],
        keywords=meta["keywords"],
        license=meta["license"],
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Topic :: Scientific/Engineering",
            "Programming Language :: Python",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
        ],
        packages=find_packages(exclude=["docs", "tests", "scripts"]),
        install_requires=get_install_requirements(),
        ext_modules=get_extensions(),
        entry_points={
            "gui_scripts": [
                "mesh_viewer = simple_3dviz.scripts.mesh_viewer:main",
                "func_viewer = simple_3dviz.scripts.func_viewer:main"
            ]
        }
    )


if __name__ == "__main__":
    setup_package()



"""Renderables represent objects that can be rendered in a scene."""

from .base import Renderable
from .lines import Lines
from .mesh import Mesh
from .spherecloud import Spherecloud
from .textured_mesh import TexturedMesh
